require_relative 'filelogger'
require 'digest/md5'

# class that handles dataset tasks report
class DatasetUtils

  DATASET_REPORT_FILE = ".dataset-task-report"
  DATASET_COMPLETED_FILE = ".dataset-completed"
  RESUME_OPERATION_FILE = ".resume-operation"
  INITIAL_INGESTION_FILE = ".initial-ingestion"

  RUN_DONE = "done"
  RUN_ERROR = "error"
  RUN_ABORTED = "aborted"
  DELETE = "DELETE"
  INGESTION = "ingestion"

  MEGA = 1000000

  def initialize(data_set_directory)
    @data_set_directory = data_set_directory
    @logger = LogManager.instance()
    @task_report_file = @data_set_directory + DATASET_REPORT_FILE
    @completed_file = @data_set_directory + DATASET_COMPLETED_FILE
    @resume_operation_file = @data_set_directory + RESUME_OPERATION_FILE
    @initial_ingestion_file = @data_set_directory + INITIAL_INGESTION_FILE
  end

  def getLocalFiles(remove=nil)
    local_files = {}
    begin
      File.readlines(@task_report_file).each do |line|
        record = line.split(";")
        if record[1].chomp == RUN_DONE
          if (remove.nil?) || (remove != record[0])
            local_files[record[0]] = {"size" => record[2].chomp, "hash" => record[3].chomp, "status" => record[1].chomp, "modification_date" => record[4].chomp }
	  end
        end
      end
    rescue Exception => e
      @logger.error("An error occurred in DatasetUtils method 'getLocalFiles':" + e.to_s)
      @logger.error(e.backtrace)
    end
    return local_files
  end

  def saveReport(local_files)
    begin
      File.delete(@task_report_file) if File.exist?(@task_report_file)
      if local_files.empty?
        File.open(@task_report_file, 'w') {}
      else
        local_files.each do |key, array|
          File.open(@task_report_file, 'ab') { |file| file.puts(key+";"+array["status"]+";"+array["size"].to_s+";"+array["hash"]+";"+array["modification_date"]) }
        end
      end
    rescue Exception => e
      @logger.error("An error occurred in DatasetUtils method 'saveReport':" + e.to_s)
      @logger.error(e.backtrace)
    end
  end

  def removeCurrentOperation()
    if File.exist?(@resume_operation_file)
      File.delete(@resume_operation_file)
    end
  end

  def saveCurrentOperation(operation, reference)
    if File.exist?(@resume_operation_file)
      File.delete(@resume_operation_file)
    end
    File.open(@resume_operation_file, 'w') { |file| file.puts(operation+";"+reference) }
  end

  def reportUpToDate(data_stream_dict)
    begin
      if not reportFileExist() and not completedFileExist()
        # directory never downloaded -new or used for partial ingestions-
        return TRUE
      end
      if reportFileExist() and not completedFileExist()
        # download not finished
        return FALSE
      end
      if data_stream_dict["status_code"] == 2
        return FALSE
      end
      if data_stream_dict["status_code"] != 0
        return TRUE
      end
      changes = getRemoteChangedDataStreams(data_stream_dict["result"])
      if changes.empty?
        return TRUE
      elsif changes.length == 1
	# check if the unique detected change corresponds to an interrumped ingestion
        if File.exist?(@resume_operation_file)
	  operation=File.open(@resume_operation_file).read.chomp.split(";")
	  if operation[0] == INGESTION
	    if operation[1] == changes[0]["reference"]
	      File.delete(@resume_operation_file)
	      return TRUE
	    end
	  end
        end
      end
      return FALSE
    rescue Exception => e
      @logger.error("An error occurred in DatasetUtils method 'reportUpToDate':" + e.to_s)
      @logger.error(e.backtrace)
      return FALSE
    end
  end

  def deleteCompletedFile()
    File.delete(@completed_file) if File.exist?(@completed_file)
  end

  def createCompletedFile()
    File.open(@completed_file, 'w') {}
  end

  def completedFileExist()
    return File.exist?(@completed_file)
  end

  def createReportFile()
    File.open(@task_report_file, 'w') {}
  end

  def reportFileExist()
    return File.exist?(@task_report_file)
  end

  def deleteInitialIngestionFile()
    File.delete(@initial_ingestion_file) if File.exist?(@initial_ingestion_file)
  end

  def createInitialIngestionFile()
    File.open(@initial_ingestion_file, 'w') {}
  end

  def initialIngestionFileExist()
    return File.exist?(@initial_ingestion_file)
  end

  def addToReport(reference, status, size, hash, data_set)
    local_files = {}
    begin
      data_set = data_set.end_with?("/") ? data_set : data_set + "/"
      file_path = @data_set_directory + reference.reverse.sub("/".reverse, ".".reverse).reverse.sub(data_set, "")
      file_path = file_path[0...-5] if file_path.end_with?(".none")
      modification_date = File.exist?(file_path) ? File.mtime(file_path).strftime("%Y-%m-%d-%H-%M-%S") : "not-modification-date"
      if not reportFileExist()
        File.open(@task_report_file, 'w') {}
      end
      new_file = TRUE
      File.readlines(@task_report_file).each do |line|
        record = line.split(";")
        if reference.to_s == record[0].to_s
          local_files[reference] = {"size" => size, "hash" => hash, "status" => status, "modification_date" => modification_date }
	  new_file = FALSE
        else
	  local_files[record[0]] = {"size" => record[2].chomp, "hash" => record[3].chomp, "status" => record[1].chomp, "modification_date" => record[4].chomp }
        end
      end
      if new_file
        local_files[reference] = {"size" => size, "hash" => hash, "status" => status, "modification_date" => modification_date }
      end
    rescue Exception => e
      @logger.error("An error occurred in DatasetUtils method 'addToReport':" + e.to_s)
      @logger.error(e.backtrace)
    end
    saveReport(local_files)
  end

  def deleteFromReport(reference, status)
    local_files = getLocalFiles(remove=reference)
    saveReport(local_files)
  end

  def getHash(file)
    begin
      chunk_size = 4 * MEGA
      md5 = Digest::MD5.new
      open(file) do |f|
        while chunk=f.read(chunk_size)
          md5.update(chunk)
        end
      end
      return md5.hexdigest
    rescue Exception => e
      @logger.error("An error occurred while getting hash of file " + file + ":" + e.to_s, print=TRUE)
      @logger.error(e.backtrace)
      raise e
    end
  end

  def getLocalChanges(files, data_set)
    all_files, new_files, modified_files, deleted_files = [], [], [], []
    begin
      if reportFileExist()
        File.readlines(@task_report_file).each do |line|
          record = line.split(";")
	  if record[1].chomp == RUN_DONE
	    data_set = data_set.end_with?("/") ? data_set : data_set + "/"
            file_path = @data_set_directory + record[0].reverse.sub("/".reverse, ".".reverse).reverse.sub(data_set, "")
            file_path = file_path[0...-5] if file_path.end_with?(".none")
            if files.include? file_path
	      modification_date = File.mtime(file_path).strftime("%Y-%m-%d-%H-%M-%S")
	      if modification_date != record[4].chomp
                size = File.size(file_path).to_s
                hash = getHash(file_path).to_s
	        if size == record[2].to_s
	          if hash != record[3].chomp
	            all_files.push({"path" => file_path, "size" => size, "hash" => hash })
		    modified_files.push(file_path)
	          end
	        else
	          all_files.push({"path" => file_path, "size" => size, "hash" => hash })
		  modified_files.push(file_path)
	        end
	      end
	      files.delete(file_path)
            else
	      all_files.push({"path" => file_path, "size" => "", "hash" => DELETE })
	      deleted_files.push(file_path)
            end
	  end
        end
      end
      files.each do |path|
        all_files.push({"path" => path, "size" => "", "hash" => "" })
	new_files.push(path)
      end
    rescue Exception => e
      @logger.error("An error occurred in DatasetUtils method 'getLocalChanges':" + e.to_s)
      @logger.error(e.backtrace)
    end
    return all_files, new_files, modified_files, deleted_files
  end

  def getRemoteChangedDataStreams(data_streams)
    pending_data_streams = []
    begin
      if reportFileExist()
	local_files = {}
	remote_files = []
	File.readlines(@task_report_file).each do |line|
	  record = line.split(";")
	  if record[1].chomp == RUN_DONE
	    local_files[record[0]] = {"size" => record[2].chomp, "hash" => record[3].chomp, }
	  end
	end
	data_streams.each do |data_stream|
	  remote_files.push(data_stream["reference"])
	  pending = TRUE
	  reference = data_stream["reference"]
	  if local_files.has_key? reference
	    size = local_files[reference]["size"]
	    if size.to_s == data_stream["size"].to_s
	      hash = local_files[reference]["hash"]
	      if hash == data_stream["hash"] or data_stream["hash"] == ""
		pending = FALSE
	      end
	    end
	  end
	  if pending
	    local_files.delete(reference)
	    pending_data_streams.push(data_stream)
	  end
	end
	local_files.each do |key, array|
	  if not remote_files.include? key
	    pending_data_streams.push({"reference" => key, "hash" => DELETE })
	  end
	end
      end
    rescue Exception => e
      @logger.error("An error occurred in DatasetUtils method 'getRemoteChangedDataStreams':" + e.to_s)
      @logger.error(e.backtrace)
    end
    return pending_data_streams
  end
end
