require_relative '../wendelin_client'
require_relative '../dataset_utils'
require_relative '../filelogger'
require 'digest/md5'

module Embulk
  module Input

    class Fifinput < InputPlugin

      Plugin.register_input("fif", self)

      NEW = "New"
      MODIFIED = "Modified"
      DELETED = "Deleted"

      EOF = "EOF"
      CHUNK_SIZE = 50000000 #50mb
      MEGA = 1000000
      SCHEMA = [
		 {"name"=>"supplier", "type"=>"string"},
		 {"name"=>"data_set", "type"=>"string"},
		 {"name"=>"file", "type"=>"string"},
		 {"name"=>"extension", "type"=>"string"},
		 {"name"=>"data_chunk", "type"=>"string"},
		 {"name"=>"eof", "type"=>"string"},
		 {"name"=>"size", "type"=>"string"},
		 {"name"=>"hash", "type"=>"string"}
               ]

      def self.showChangesList(changes, type, print_short)
	if not changes.empty?
	  puts
	  @logger.info("#{type} file(s):", print=TRUE)
	  if print_short and changes.length > 50
	    limit = changes.length > 130 ? 130/3 : changes.length/3
	    @logger.info(changes[0, limit], print=TRUE)
	    @logger.info("....", print=TRUE)
	    @logger.info(changes[changes.length-limit, changes.length-1], print=TRUE)
	  else
	    @logger.info(changes, print=TRUE)
	  end
	end
      end

      def self.transaction(config, &control)
	begin
	  tool_dir = config.param('tool_dir', :string)
	  @logger = LogManager.instance()
	  @logger.setFilename(tool_dir, "ingestion")
	  task = { 'paths' => [] }
	  task['supplier'] = config.param('supplier', :string)
	  task['data_set'] = config.param('data_set', :string)
	  task['chunk_size'] = config.param('chunk_size', :float, default: 0) * MEGA
	  if task['chunk_size'] == 0
	    task['chunk_size'] = CHUNK_SIZE
	  end
	  @data_set = task['data_set']
	  paths = config.param('path_prefix', :array)
	  paths[0] = paths[0].end_with?("/") ? paths[0] : paths[0] + "/"
	  @data_set_directory = paths[0]
	  task['inputs'] = paths
	  @logger.info("Getting local files for ingestion...", print=TRUE)
	  task['paths'] = paths.map {|path|
				      next [] unless Dir.exist?(path)
				      Dir[(path + '/**/*').gsub! '//', '/']
				    }.flatten.select{ |file| File.file?(file) }

	  @wendelin = WendelinClient.new(config.param('erp5_url', :string), config.param('user', :string), config.param('password', :string))
	  @logger.info("Checking remote dataset...", print=TRUE)
	  data_stream_dict = @wendelin.getDataStreams(task['data_set'])
	  @dataset_utils = DatasetUtils.new(@data_set_directory)
	  if not @dataset_utils.reportFileExist()
	    @dataset_utils.createInitialIngestionFile()
	  else
	    if not @dataset_utils.initialIngestionFileExist()
	      @logger.info("Checking local dataset...", print=TRUE)
	      if not @dataset_utils.reportUpToDate(data_stream_dict)
	        puts
	        @logger.error("Your current dataset is outdated. Please, run a download to update it before ingest your changes.", print=TRUE)
	        puts
	        @logger.abortExecution(error=FALSE)
	      end
	    end
	  end
	  if data_stream_dict["status_code"] != 0
	    @logger.error(data_stream_dict["error_message"], print=TRUE)
	    @logger.abortExecution()
          end

	  @logger.info("Supplier: #{task['supplier']}")
	  @logger.info("Dataset name: #{task['data_set']}")
	  @logger.info("Chunk size set in #{task['chunk_size']/MEGA}MB")
	  if task['data_set'] == '$DATA_SET'
	    @logger.error("There was an error setting the configuration file", print=TRUE)
            @logger.info("Please try manual ingestion or update manually the ingestion configuration file.", print=TRUE)
	    @logger.abortExecution()
	  end
	  if task['paths'].empty? and not @dataset_utils.reportFileExist()
	    @logger.error("The dataset directory '#{task['inputs'][0]}' is empty.", print=TRUE)
	    @logger.error("Could not find any valid file.", print=TRUE)
	    @logger.error("Please make sure your dataset directory contains files for ingestion.", print=TRUE)
	    @logger.abortExecution()
          end

	  task['paths'], new_files, modified_files, deleted_files = @dataset_utils.getLocalChanges(task['paths'], task['data_set'])
	  if task['paths'].empty?
	    puts
	    @logger.info("No changes in '#{@data_set_directory}'. Everything up-to-date.", print=TRUE)
	    @logger.abortExecution(error=FALSE)
	  end
	  changes = @dataset_utils.reportFileExist() ? "change" : "new file"
	  @logger.info("#{task['paths'].length} #{changes}(s) detected for ingestion: ", print=TRUE)
	  print_short = task['paths'].length > 500
	  self.showChangesList(new_files, NEW, print_short)
	  self.showChangesList(modified_files, MODIFIED, print_short)
	  self.showChangesList(deleted_files, DELETED, print_short)
	  puts
	  @logger.info("Continue with ingestion? (y/n)", print=TRUE)
          option = gets
          option = option.chomp
	  if option == "n"
	    @logger.info("Ingestion cancelled by user.", print=TRUE)
	    @logger.abortExecution()
	  end
	  if not @dataset_utils.reportFileExist()
	    @dataset_utils.createReportFile()
	  end

	  columns = [
            Column.new(0, "supplier", :string),
            Column.new(1, "data_set", :string),
	    Column.new(2, "file", :string),
	    Column.new(3, "extension", :string),
	    Column.new(4, "data_chunk", :string),
	    Column.new(5, "eof", :string),
	    Column.new(6, "size", :string),
	    Column.new(7, "hash", :string)
          ]

	  commit_reports = yield(task, columns, task['paths'].length)
          done = commit_reports.map{|hash| hash["done"]}.flatten.compact

          resume(task, columns, task['paths'].length, &control)
	rescue Exception => e
	  @logger.error("An error occurred during operation: " + e.to_s, print=TRUE)
	  @logger.error(e.backtrace)
	  puts "[INFO] For more detailed information, please refer to the log file."
	  @logger.abortExecution()
	end
      end

      def self.resume(task, columns, count, &control)
	@logger = LogManager.instance()
        task_reports = yield(task, columns, count)
	@logger.info("Reports:", print=TRUE)
	if task_reports.length > 15
	  @logger.info(task_reports[0, 5], print=TRUE)
	  @logger.info(".....", print=TRUE)
	  @logger.info(task_reports[task_reports.length-5, task_reports.length-1], print=TRUE)
	  @logger.info("Full task report:")
	  @logger.info(task_reports)
	else
	  @logger.info(task_reports, print=TRUE)
	end
        next_config_diff = task_reports.map{|hash| hash["done"]}.flatten.compact
	changes = @dataset_utils.initialIngestionFileExist() ? "new file" : "change"
	@logger.info("#{next_config_diff.length} #{changes}(s) ingested.", print=TRUE)
	if(next_config_diff.length == count)
	  @logger.info("Dataset successfully ingested.", print=TRUE)
	  @wendelin.increaseDatasetVersion(@data_set)
	else
	  next_config_diff = task_reports.map{|hash| hash["error"]}.flatten.compact
	  puts
	  @logger.error("The following files could not be ingested. Please check the details in the log file: " + @logger.getLogPath(), print=TRUE)
	  if next_config_diff.length > 15
	    @logger.error(next_config_diff[0, 5], print=TRUE)
	    @logger.error(".....", print=TRUE)
	    @logger.error(next_config_diff[next_config_diff.length-5, next_config_diff.length-1], print=TRUE)
	  else
	    @logger.error(next_config_diff, print=TRUE)
	  end
	  @logger.info("You can retry the ingestion for those pending files.", print=TRUE)
	  puts
	end
	next_config_diff = {}
	return {"done" => next_config_diff}        
      end

      def initialize(task, schema, index, page_builder)
        super
        @supplier = task['supplier']
        @dataset = task['data_set']
	@chunk_size = task['chunk_size']
	@input_dirs = task['inputs']
	@data_set_directory = task['inputs'][0]
	@logger = LogManager.instance()
	@dataset_utils = DatasetUtils.new(@data_set_directory)
      end

      def run
	file_dict = task['paths'][@index]
	@logger.info("Processing file #{file_dict["path"].to_s}", print=TRUE)
	begin
	  path = file_dict["path"]
	  size = file_dict["size"]
	  hash = file_dict["hash"]
	  delete = hash == DatasetUtils::DELETE
	  if size == "" and hash == "" #new file
	    size = File.size(path)
	    hash = @dataset_utils.getHash(path)
	  end
	  extension = File.extname path
	  if path.start_with?(@input_dirs[0])
	    filename = path.sub(@input_dirs[0], "")
	    filename = filename.reverse.sub(extension.reverse, "").reverse
	  end
	  extension.gsub! '.', ''
	  extension = extension == "" ? "none" : extension

	  dataset_file = file_dict["path"].sub(@data_set_directory, @dataset.end_with?("/") ? @dataset : @dataset + "/")
	  if extension != "none"
	    old_pattern = filename + '.' + extension
	    new_pattern = filename + '/' + extension
	    dataset_file = dataset_file.reverse.sub(old_pattern.reverse, new_pattern.reverse).reverse
	  else
	    dataset_file += "/" + extension
	  end
	  @dataset_utils.saveCurrentOperation(DatasetUtils::INGESTION, dataset_file)
	  each_chunk(path, filename, extension, size, hash, schema[1..-1].map{|elm| elm.name}, @chunk_size, delete) do |entry|
	    @page_builder.add(entry)
          end
          @page_builder.finish
	rescue java.lang.OutOfMemoryError
	  @logger.logOutOfMemoryError(file_dict["path"])
	  return_value = DatasetUtils::RUN_ABORTED
	rescue Exception => e
	  @logger.error("An error occurred during file ingestion: " + e.to_s, print=TRUE)
	  @logger.error(e.backtrace)
	  puts "[INFO] For more detailed information, please refer to the log file: " + @logger.getLogPath()
	  return_value = DatasetUtils::RUN_ERROR
	else
	  return_value = DatasetUtils::RUN_DONE
	end

	# update reports if operation successfully ended
	if return_value == DatasetUtils::RUN_DONE
	  if delete
	    if @dataset_utils.reportFileExist()
	      @dataset_utils.deleteFromReport(dataset_file, return_value)
	    end
	  else
	    if @dataset_utils.reportFileExist()
	      @dataset_utils.addToReport(dataset_file, return_value, size, hash, task['data_set'])
	    end
	  end
	end

	@dataset_utils.removeCurrentOperation()
	return {return_value => file_dict["path"]}
      end

      private

      def each_chunk(path, filename, extension, size, hash, fields, chunk_size=CHUNK_SIZE, delete=FALSE)
	if delete
	  values = [@supplier, @dataset, filename, extension, "", DatasetUtils::DELETE, "", ""]
	  yield(values)
	else
          file_object = File.open(path, "rb")
	  npart = 0
          next_byte = file_object.read(1)
	  first = TRUE
          while true
              data = next_byte
              if not next_byte
		  if first # this means this is an empty file
		    values = [@supplier, @dataset, filename, extension, "", "", size, hash]
		    yield(values)
		  end
		  break
	      end
              data += file_object.read(chunk_size)
              next_byte = file_object.read(1)
              if not next_byte
                  eof = EOF
		  if first # this means that the whole file will be ingested at once (not split)
		    eof = ""
		  end
	      else
		  npart += 1
       	 	  eof = npart.to_s.rjust(3, "0")
	      end
	      content = Base64.encode64(data)
	      values = [@supplier, @dataset, filename, extension, content, eof, size, hash]
	      first = FALSE
	      yield(values)
	  end
	end
      end
    end
  end
end
