require_relative '../wendelin_client'
require_relative '../dataset_utils'
require 'fileutils'

module Embulk
  module Input

    class Wendelininput < InputPlugin

      CHUNK_SIZE = 50000000 #50mb
      MEGA = 1000000

      UPDATE = "U"
      RESUME = "R"
      DOWNLOAD = "D"
      ABORT = "A"

      Plugin.register_input("wendelin", self)

      def self.warnConflicts(remote_streams, data_set, action)
	if not remote_streams.empty?
	  paths = [@data_set_directory.end_with?("/") ? @data_set_directory : @data_set_directory + "/"]
	  local_files = paths.map {|path|
				      next [] unless Dir.exist?(path)
				      Dir[(path + '/**/*').gsub! '//', '/']
				    }.flatten.select{ |file| File.file?(file) }
	  local_changes, a, b, c = @dataset_utils.getLocalChanges(local_files, data_set)
	  data_set = @data_set.end_with?("/") ? @data_set : @data_set + "/"
	  remote_changes = remote_streams.map { |remote|
		remote = @data_set_directory + remote["reference"].reverse.sub("/".reverse, ".".reverse).reverse.sub(data_set, "")
		remote.end_with?(".none") ? remote[0...-5] : remote
	  }
	  conflicts = local_changes.select{ |conflict| remote_changes.include? conflict["path"] }.map{ |conflict| conflict["path"] }
	  # check scenario where the last version file exists but not in report
	  # (due download interrumped right after save the file but before add it to report)
	  if action == RESUME and conflicts.length == 1 and File.exist?(conflicts[0])
	    @logger.warn("The file #{conflicts[0]} was detected as false positive conflict and it was not informed to user.")
	    conflicts = []
	  end
	  if not conflicts.empty?
	    @logger.warn("CONFLICT: there are conflicts with some of your local changes.", print=TRUE)
	    puts "** press key **"
	    option = gets
	    @logger.warn("Conflicted files:", print=TRUE)
	    @logger.warn(conflicts, print=TRUE)
	    puts
	    @logger.warn("Your local conflicted files will be overwritten by download.", print=TRUE)
	    @logger.warn("Do you want to continue? (y/n)", print=TRUE)
	    option = gets
	    option = option.chomp
	    if option == "n" or option == "N"
	      @logger.info("Download cancelled by user.", print=TRUE)
	      @logger.abortExecution(error=FALSE)
	    end
	  end
	end
      end

      def self.askUserForAction(task, action)
	    if action == RESUME
	      action_message =  "#{RESUME}: Resume. Continues download from last file."
	    else
	      action = UPDATE
	      action_message = "#{UPDATE}: Update. Checks for changes in dataset."
            end
	    valid_option = FALSE
	    while not valid_option
	      @logger.info("Please select an option [#{action}, #{DOWNLOAD}, #{ABORT}]", print=TRUE)
	      @logger.info(action_message, print=TRUE)
	      @logger.info("#{DOWNLOAD}: Download. Downloads the dataset from scratch.", print=TRUE)
	      @logger.info("#{ABORT}: Abort operation.", print=TRUE)
	      option = gets
	      option = option.chomp
	      if not [action, DOWNLOAD, ABORT].include? option
		@logger.info("Invalid option", print=TRUE)
	      else
		valid_option = TRUE
	      end
	    end
	    case option
	      when action
		task['data_streams'] = @dataset_utils.getRemoteChangedDataStreams(task['data_streams'])
		self.warnConflicts(task['data_streams'], task['data_set'], action)
		@dataset_utils.deleteCompletedFile()
	        if task['data_streams'].empty?
	          @logger.info("No new files in dataset.", print=TRUE)
	          @logger.info("Your downloaded dataset is already up to date.", print=TRUE)
	        end
	      when DOWNLOAD
		ebulk_file = @data_set_directory + "/.ebulk_dataset"
		ebulk_file_content = ""
		if File.file?(ebulk_file)
		  ebulk_file_content = File.read(ebulk_file)
		end
		FileUtils.rm_rf(@data_set_directory)
		unless File.directory?(@data_set_directory)
		  FileUtils.mkdir_p(@data_set_directory)
		end
		if ebulk_file_content != ""
		  File.open(ebulk_file, 'w') { |file| file.write(ebulk_file_content) }
		end
		@dataset_utils.deleteCompletedFile()
		@dataset_utils.createReportFile()
	      when ABORT
		@logger.abortExecution()
	    end
      end

      def self.transaction(config, &control)
	begin
	  @tool_dir = config.param('tool_dir', :string)
	  @logger = LogManager.instance()
	  @logger.setFilename(@tool_dir, "download")
	  @erp5_url = config.param('erp5_url', :string)
	  @data_set = config.param('data_set', :string)
	  @logger.info("Dataset name: #{@data_set}")
	  if @data_set == '$DATA_SET'
	    @logger.error("There was an error setting the configuration file", print=TRUE)
	    @logger.info("Please try manual download or update manually the download configuration file.", print=TRUE)
	    @logger.abortExecution()
	  end
	  @user = config.param("user", :string, defualt: nil)
	  @logger.info("User: #{@user}")
	  @password = config.param("password", :string, default: nil)
	  @chunk_size = config.param('chunk_size', :float, default: 0) * MEGA
	  @output_path = config.param("output_path", :string,  :default => nil)
	  if File.directory?(@output_path)
	  else
	    @logger.error("Output directory not found.", print=TRUE)
	    @logger.abortExecution()
	  end
	  @wendelin = WendelinClient.new(@erp5_url, @user, @password)

	  task = {
		 'erp5_url' => @erp5_url,
		 'data_set' => @data_set,
		 'user' => @user,
		 'password' => @password,
		 'chunk_size' => @chunk_size,
		 'output_path' => @output_path,
		 'tool_dir' => @tool_dir
	       }
	  if task['chunk_size'] == 0
	    task['chunk_size'] = CHUNK_SIZE
	  end
	  @logger.info("Chunk size set in #{task['chunk_size']/MEGA}MB")
	  @data_set_directory = @output_path.end_with?("/") ? @output_path : @output_path + "/"
	  task['data_set_directory'] = @data_set_directory
	  @dataset_utils = DatasetUtils.new(@data_set_directory)

	  @logger.info("Getting remote file list from dataset '#{@data_set}'...", print=TRUE)
	  data_stream_list = @wendelin.getDataStreams(@data_set)
	  n_retry = 0
	  while data_stream_list["status_code"] == 2 and n_retry < 6
	    sleep 10
	    data_stream_list = @wendelin.getDataStreams(@data_set)
	    n_retry += 1
	  end
	  if data_stream_list["status_code"] == 0
	    if data_stream_list["result"].empty?
	      @logger.error("No valid data found for data set " + @data_set, print=TRUE)
	      @logger.abortExecution(error=FALSE)
	    end
	    task['data_streams'] = data_stream_list["result"]
	  elsif data_stream_list["status_code"] == 2
	    @logger.error("Dataset '#{@data_set}' has files recently ingested waiting for processing.", print=TRUE)
	    @logger.error("Please retry in some minutes.", print=TRUE)
	    @logger.abortExecution(error=FALSE)
	  else
	    @logger.error(data_stream_list["error_message"], print=TRUE)
	    @logger.abortExecution()
	  end
	  @logger.info("Done", print=TRUE)

	  if @dataset_utils.reportFileExist()
	    if @dataset_utils.completedFileExist()
	      puts
	      @logger.info("This dataset was already downloaded. What do you want to do?", print=TRUE)
	      puts
	      self.askUserForAction(task, action=UPDATE)
	    elsif not @dataset_utils.initialIngestionFileExist()
	      puts
	      @logger.info("There was a previous attempt to download this dataset but it did not finish successfully.", print=TRUE)
	      @logger.info("What do you want to do?", print=TRUE)
	      puts
	      self.askUserForAction(task, action=RESUME)
	    else
	      puts
	      self.askUserForAction(task, action=UPDATE)
	    end
	  else
	    dir_entries = Dir.entries(@data_set_directory).length
	    if File.file?(@data_set_directory+"/.ebulk_dataset")
	      dir_entries -= 1
	    end
	    if dir_entries > 2
	      puts
	      @logger.info("Dataset download directory is not empty! It will be overwritten: " + @data_set_directory, print=TRUE)
	      @logger.info("Continue with download? (y/n)", print=TRUE)
	      option = gets
	      option = option.chomp
	      if option == "n"
	        @logger.info("Download cancelled by user.", print=TRUE)
	        @logger.abortExecution(error=FALSE)
	      end
	    end
	    @dataset_utils.createReportFile()
	  end
	  @dataset_utils.deleteInitialIngestionFile()
	  columns = [
	    Column.new(0, "reference", :string),
	    Column.new(1, "data_chunk", :string),
	    Column.new(2, "data_set", :string),
	    Column.new(3, "mode", :string)
	  ]
	  resume(task, columns, task['data_streams'].length, &control)
	rescue Exception => e
	  @logger.error("An error occurred during operation: " + e.to_s, print=TRUE)
	  @logger.error(e.backtrace)
	  puts "[INFO] For more detailed information, please refer to the log file."
	  @logger.abortExecution()
	end
      end

      def self.resume(task, columns, count, &control)
	@logger = LogManager.instance()
        task_reports = yield(task, columns, count)
	if task_reports.any?
	  @logger.info("Reports:", print=TRUE)
	  if task_reports.length > 15
	    @logger.info(task_reports[0, 5], print=TRUE)
	    @logger.info(".....", print=TRUE)
	    @logger.info(task_reports[task_reports.length-5, task_reports.length-1], print=TRUE)
	  else
	    @logger.info(task_reports, print=TRUE)
	  end
	  @logger.info("Full task report:")
	  @logger.info(task_reports)
	end
        next_config_diff = task_reports.map{|hash| hash[DatasetUtils::RUN_DONE]}.flatten.compact
	if(next_config_diff.length == count)
	  if(count > 0)
	    @logger.info("Dataset successfully downloaded.", print=TRUE)
	    @logger.info("#{count} files processed.", print=TRUE)
	    @logger.info("Dataset files are in dataset directory: " + @data_set_directory, print=TRUE)
	  end
	  @dataset_utils.createCompletedFile()
	  if count > 10
	    next_config_diff = {}
	  end
	end
	return {DatasetUtils::RUN_DONE => next_config_diff}
      end

      def initialize(task, schema, index, page_builder)
        super
        @data_set = task['data_set']
	@chunk_size = task['chunk_size']
	@data_set_directory = task['data_set_directory']
	@wendelin = WendelinClient.new(task['erp5_url'], task['user'], task['password'])
	@logger = LogManager.instance()
	@dataset_utils = DatasetUtils.new(@data_set_directory)
      end

      def run
	data_stream = task['data_streams'][@index]
	id = data_stream["id"]
	ref = data_stream["reference"]
	size = data_stream["size"]
	hash = data_stream["hash"]
	begin
	  if hash.to_s == DatasetUtils::DELETE
	    @logger.info("Deleting #{ref}", print=TRUE)
	    entry = [ref, "", @data_set, hash.to_s]
  	    page_builder.add(entry)
	  else
	    @logger.info("Getting content from remote #{ref}", print=TRUE)
	    n_chunk = 0
	    @wendelin.eachDataStreamContentChunk(id, @chunk_size) do |chunk|
	      content = chunk.nil? || chunk.empty? ? "" : Base64.encode64(chunk)
	      begin_of_file = n_chunk == 0
	      entry = [ref, content, @data_set, begin_of_file]
	      page_builder.add(entry)
	      n_chunk += 1
	    end
	  end
          page_builder.finish
	rescue java.lang.OutOfMemoryError
          @logger.logOutOfMemoryError(ref)
	  return_value = DatasetUtils::RUN_ABORTED
        rescue Exception => e  
	  @logger.error(e.to_s, print=TRUE)
	  @logger.error(e.backtrace)
	  puts "[INFO] For more detailed information, please refer to the log file: " + @logger.getLogPath()
	  return_value = DatasetUtils::RUN_ERROR
	else
	  return_value = DatasetUtils::RUN_DONE
	end
	# update reports if operation successfully ended
	if return_value == DatasetUtils::RUN_DONE
	  if hash.to_s == DatasetUtils::DELETE
	    @dataset_utils.deleteFromReport(ref, return_value)
	  else
	    @dataset_utils.addToReport(ref, return_value, size, hash, task['data_set'])
	  end
	end
        return {return_value => ref}
      end
    end
  end
end
